@@ -1,14 +1,51 @@
 ### Terrence D. Jorgensen
-### Last updated: 26 February 2016
+### Last updated: 25 June 2018
 ### Savalei & Yuan's (2009) model-based bootstrap for missing data
 
-setClass("BootMiss", representation(time = "list", transData = "data.frame", bootDist = "vector", origChi = "numeric", df = "numeric", bootP="numeric"))
 
+## ----------------------------
+## "BootMiss" Class and Methods
+## ----------------------------
 
-#########################################
-## Define methods for class "BootMiss" ##
-#########################################
+#' Class For the Results of Bollen-Stine Bootstrap with Incomplete Data
+#'
+#' This class contains the results of Bollen-Stine bootstrap with missing data.
+#'
+#'
+#' @name BootMiss-class
+#' @aliases BootMiss-class show,BootMiss-method summary,BootMiss-method
+#' hist,BootMiss-method
+#' @docType class
+#' @section Objects from the Class: Objects can be created via the
+#' \code{\link{bsBootMiss}} function.
+#' @slot time A list containing 2 \code{difftime} objects (\code{transform}
+#'  and \code{fit}), indicating the time elapsed for data transformation and
+#'  for fitting the model to bootstrap data sets, respectively.
+#' @slot transData Transformed data
+#' @slot bootDist The vector of \eqn{chi^2} values from bootstrap data sets
+#'  fitted by the target model
+#' @slot origChi The \eqn{chi^2} value from the original data set
+#' @slot df The degree of freedom of the model
+#' @slot bootP The \emph{p} value comparing the original \eqn{chi^2} with the
+#'  bootstrap distribution
+#' @author Terrence D. Jorgensen (University of Amsterdam;
+#' \email{TJorgensen314@@gmail.com})
+#' @seealso \code{\link{bsBootMiss}}
+#' @examples
+#'
+#' # See the example from the bsBootMiss function
+#'
+setClass("BootMiss", representation(time = "list",
+                                    transData = "data.frame",
+                                    bootDist = "vector",
+                                    origChi = "numeric",
+                                    df = "numeric",
+                                    bootP = "numeric"))
 
+#' @rdname BootMiss-class
+#' @aliases show,BootMiss-method
+#' @importFrom stats pchisq
+#' @export
 setMethod("show", "BootMiss",
 function(object) {
   cat("Chi-Squared = ", object@origChi, "\nDegrees of Freedom = ",
@@ -20,6 +57,10 @@ function(object) {
   invisible(object)
 })
 
+#' @rdname BootMiss-class
+#' @aliases summary,BootMiss-method
+#' @importFrom stats var
+#' @export
 setMethod("summary", "BootMiss",
 function(object) {
   cat("Time elapsed to transform the data:\n")
@@ -36,6 +77,22 @@ function(object) {
   invisible(object)
 })
 
+#' @rdname BootMiss-class
+#' @aliases hist,BootMiss-method
+#' @importFrom stats qchisq dchisq quantile
+#' @param object,x object of class \code{BootMiss}
+#' @param ... Additional arguments to pass to \code{\link[graphics]{hist}}
+#' @param alpha alpha level used to draw confidence limits
+#' @param nd number of digits to display
+#' @param printLegend \code{logical}. If \code{TRUE} (default), a legend will
+#'  be printed with the histogram
+#' @param legendArgs \code{list} of arguments passed to the
+#'  \code{\link[graphics]{legend}} function.  The default argument is a list
+#'  placing the legend at the top-left of the figure.
+#' @return The \code{hist} method returns a list of \code{length == 2},
+#'  containing the arguments for the call to \code{hist} and the arguments
+#'  to the call for \code{legend}, respectively.
+#' @export
 setMethod("hist", "BootMiss",
 function(x, ..., alpha = .05, nd = 2, printLegend = TRUE,
          legendArgs = list(x = "topleft")) {
@@ -98,6 +155,403 @@ function(x, ..., alpha = .05, nd = 2, printLegend = TRUE,
   invisible(list(hist = histArgs, legend = legendArgs))
 })
 
+
+
+## --------------------
+## Constructor Function
+## --------------------
+
+#' Bollen-Stine Bootstrap with the Existence of Missing Data
+#'
+#' Implement the Bollen and Stine's (1992) Bootstrap when missing observations
+#' exist. The implemented method is proposed by Savalei and Yuan (2009). This
+#' can be used in two ways. The first and easiest option is to fit the model to
+#' incomplete data in \code{lavaan} using the FIML estimator, then pass that
+#' \code{lavaan} object to \code{bsBootMiss}.
+#'
+#' The second is designed for users of other software packages (e.g., LISREL,
+#' EQS, Amos, or Mplus). Users can import their data, \eqn{\chi^2} value, and
+#' model-implied moments from another package, and they have the option of
+#' saving (or writing to a file) either the transformed data or bootstrapped
+#' samples of that data, which can be analyzed in other programs. In order to
+#' analyze the bootstrapped samples and return a \emph{p} value, users of other
+#' programs must still specify their model using lavaan syntax.
+#'
+#'
+#' @importFrom lavaan lavInspect parTable
+#' @param x A target \code{lavaan} object used in the Bollen-Stine bootstrap
+#' @param transformation The transformation methods in Savalei and Yuan (2009).
+#' There are three methods in the article, but only the first two are currently
+#' implemented here.  Use \code{transformation = 1} when there are few missing
+#' data patterns, each of which has a large size, such as in a
+#' planned-missing-data design.  Use \code{transformation = 2} when there are
+#' more missing data patterns. The currently unavailable
+#' \code{transformation = 3} would be used when several missing data patterns
+#' have n = 1.
+#' @param nBoot The number of bootstrap samples.
+#' @param model Optional. The target model if \code{x} is not provided.
+#' @param rawData Optional. The target raw data set if \code{x} is not
+#'  provided.
+#' @param Sigma Optional. The model-implied covariance matrix if \code{x} is
+#'  not provided.
+#' @param Mu Optional. The model-implied mean vector if \code{x} is not
+#'  provided.
+#' @param group Optional character string specifying the name of the grouping
+#'  variable in \code{rawData} if \code{x} is not provided.
+#' @param ChiSquared Optional. The model's \eqn{\chi^2} test statistic if
+#'  \code{x} is not provided.
+#' @param EMcov Optional, if \code{x} is not provided. The EM (or Two-Stage ML)
+#' estimated covariance matrix used to speed up Transformation 2 algorithm.
+#' @param transDataOnly Logical. If \code{TRUE}, the result will provide the
+#' transformed data only.
+#' @param writeTransData Logical. If \code{TRUE}, the transformed data set is
+#' written to a text file, \code{transDataOnly} is set to \code{TRUE}, and the
+#' transformed data is returned invisibly.
+#' @param bootSamplesOnly Logical. If \code{TRUE}, the result will provide
+#' bootstrap data sets only.
+#' @param writeBootData Logical. If \code{TRUE}, the stacked bootstrap data
+#' sets are written to a text file, \code{bootSamplesOnly} is set to
+#' \code{TRUE}, and the list of bootstrap data sets are returned invisibly.
+#' @param writeArgs Optional \code{list}. If \code{writeBootData = TRUE} or
+#' \code{writeBootData = TRUE}, user can pass arguments to the
+#' \code{\link[utils]{write.table}} function as a list.  Some default values
+#' are provided: \code{file} = "bootstrappedSamples.dat", \code{row.names} =
+#' \code{FALSE}, and \code{na} = "-999", but the user can override all of these
+#' by providing other values for those arguments in the \code{writeArgs} list.
+#' @param seed The seed number used in randomly drawing bootstrap samples.
+#' @param suppressWarn Logical. If \code{TRUE}, warnings from \code{lavaan}
+#' function will be suppressed when fitting the model to each bootstrap sample.
+#' @param showProgress Logical. Indicating whether to display a progress bar
+#' while fitting models to bootstrap samples.
+#' @param \dots The additional arguments in the \code{\link[lavaan]{lavaan}}
+#' function. See also \code{\link[lavaan]{lavOptions}}
+#' @return As a default, this function returns a \code{\linkS4class{BootMiss}}
+#' object containing the results of the bootstrap samples. Use \code{show},
+#' \code{summary}, or \code{hist} to examine the results. Optionally, the
+#' transformed data set is returned if \code{transDataOnly = TRUE}. Optionally,
+#' the bootstrap data sets are returned if \code{bootSamplesOnly = TRUE}.
+#' @author Terrence D. Jorgensen (University of Amsterdam;
+#' \email{TJorgensen314@@gmail.com})
+#'
+#' Syntax for transformations borrowed from http://www2.psych.ubc.ca/~vsavalei/
+#' @seealso \code{\linkS4class{BootMiss}}
+#' @references
+#'
+#' Bollen, K. A., & Stine, R. A. (1992). Bootstrapping goodness-of-fit measures
+#' in structural equation models. \emph{Sociological Methods &
+#' Research, 21}(2), 205--229. doi:10.1177/0049124192021002004
+#'
+#' Savalei, V., & Yuan, K.-H. (2009). On the model-based bootstrap with missing
+#' data: Obtaining a p-value for a test of exact fit. \emph{Multivariate
+#' Behavioral Research, 44}(6), 741--763. doi:10.1080/00273170903333590
+#' @examples
+#'
+#' \dontrun{
+#' dat1 <- HolzingerSwineford1939
+#' dat1$x5 <- ifelse(dat1$x1 <= quantile(dat1$x1, .3), NA, dat1$x5)
+#' dat1$x9 <- ifelse(is.na(dat1$x5), NA, dat1$x9)
+#'
+#' targetModel <- "
+#' visual  =~ x1 + x2 + x3
+#' textual =~ x4 + x5 + x6
+#' speed   =~ x7 + x8 + x9
+#' "
+#' targetFit <- sem(targetModel, dat1, meanstructure = TRUE, std.lv = TRUE,
+#'                  missing = "fiml", group = "school")
+#' summary(targetFit, fit = TRUE, standardized = TRUE)
+#'
+#' # The number of bootstrap samples should be much higher.
+#' temp <- bsBootMiss(targetFit, transformation = 1, nBoot = 10, seed = 31415)
+#'
+#' temp
+#' summary(temp)
+#' hist(temp)
+#' hist(temp, printLegend = FALSE) # suppress the legend
+#' ## user can specify alpha level (default: alpha = 0.05), and the number of
+#' ## digits to display (default: nd = 2).  Pass other arguments to hist(...),
+#' ## or a list of arguments to legend() via "legendArgs"
+#' hist(temp, alpha = .01, nd = 3, xlab = "something else", breaks = 25,
+#'      legendArgs = list("bottomleft", box.lty = 2))
+#' }
+#'
+#' @export
+bsBootMiss <- function(x, transformation = 2, nBoot = 500, model, rawData,
+                       Sigma, Mu, group, ChiSquared, EMcov,
+                       writeTransData = FALSE, transDataOnly = FALSE,
+                       writeBootData = FALSE, bootSamplesOnly = FALSE,
+                       writeArgs, seed = NULL, suppressWarn = TRUE,
+                       showProgress = TRUE, ...) {
+  if(writeTransData) transDataOnly <- TRUE
+  if(writeBootData) bootSamplesOnly <- TRUE
+
+  check.nBoot <- (!is.numeric(nBoot) | nBoot < 1L) & !transDataOnly
+  if (check.nBoot) stop("The \"nBoot\" argument must be a positive integer.")
+
+  ## Which transformation?
+  if (!(transformation %in% 1:2)) stop("User must specify transformation 1 or 2.
+       Consult Savalei & Yuan (2009) for advice.
+       Transformation 3 is not currently available.")
+  if (transformation == 2) SavaleiYuan <- trans2
+  #if (transformation == 3) SavaleiYuan <- trans3
+
+  ######################
+  ## Data Preparation ##
+  ######################
+
+  ## If a lavaan object is supplied, the extracted values for rawData, Sigma, Mu,
+  ## EMcov, and EMmeans will override any user-supplied arguments.
+  nG <- lavInspect(x, "ngroups")
+  if (hasArg(x)) {
+    if (nG == 1L) {
+      rawData <- list(as.data.frame(lavInspect(x, "data")))
+    } else rawData <- lapply(lavInspect(x, "data"), as.data.frame)
+    for (g in seq_along(rawData)) {
+      colnames(rawData[[g]]) <- lavaan::lavNames(x)
+      checkAllMissing <- apply(rawData[[g]], 1, function(x) all(is.na(x)))
+      if (any(checkAllMissing)) rawData[[g]] <- rawData[[g]][!checkAllMissing, ]
+    }
+    ChiSquared <- lavInspect(x, "fit")[c("chisq", "chisq.scaled")]
+    ChiSquared <- ifelse(is.na(ChiSquared[2]), ChiSquared[1], ChiSquared[2])
+    group <- lavInspect(x, "group")
+    if (length(group) == 0) group <- "group"
+    group.label <- lavInspect(x, "group.label")
+    if (length(group.label) == 0) group.label <- 1
+    Sigma <- lavInspect(x, "cov.ov")
+    Mu <- lavInspect(x, "mean.ov")
+    EMcov <- lavInspect(x, "sampstat")$cov
+    if (nG == 1L) {
+      Sigma <- list(Sigma)
+      Mu <- list(Mu)
+      EMcov <- list(EMcov)
+    }
+  } else {
+  ## If no lavaan object is supplied, check that required arguments are.
+    suppliedData <- c(hasArg(rawData), hasArg(Sigma), hasArg(Mu))
+    if (!all(suppliedData)) {
+      stop("Without a lavaan fitted object, user must supply raw data and",
+           " model-implied moments.")
+    }
+    if (!hasArg(model) & !(transDataOnly | bootSamplesOnly)) {
+      stop("Without model syntax or fitted lavaan object, user can only call",
+           " this function to save transformed data or bootstrapped samples.")
+    }
+    if (!hasArg(ChiSquared) & !(transDataOnly | bootSamplesOnly)) {
+      stop("Without a fitted lavaan object or ChiSquared argument, user can",
+           " only call this function to save transformed data, bootstrapped",
+           " samples, or bootstrapped chi-squared values.")
+    }
+    if (!any(c(transDataOnly, bootSamplesOnly))) {
+      if (!is.numeric(ChiSquared)) stop("The \"ChiSquared\" argument must be numeric.")
+    }
+
+    ## If user supplies one-group data & moments, convert to lists.
+    if (class(rawData) == "data.frame") rawData <- list(rawData)
+    if (class(rawData) != "list") {
+      stop("The \"rawData\" argument must be a data.frame or list of data frames.")
+    } else {
+      if (!all(sapply(rawData, is.data.frame))) stop("Every element of \"rawData\" must be a data.frame")
+    }
+    if (class(Sigma) == "matrix") Sigma <- list(Sigma)
+    if (is.numeric(Mu)) Mu <- list(Mu)
+
+    ## check whether EMcov was supplied for starting values in Trans2/Trans3
+    if (!hasArg(EMcov)) {
+      EMcov <- vector("list", length(Sigma))
+    } else {
+      if (class(EMcov) == "matrix") EMcov <- list(EMcov)
+      ## check EMcov is symmetric and dimensions match Sigma
+      for (g in seq_along(EMcov)) {
+        if (!isSymmetric(EMcov[[g]])) stop("EMcov in group ", g, " not symmetric.")
+        unequalDim <- !all(dim(EMcov[[g]]) == dim(Sigma[[g]]))
+        if (unequalDim) stop("Unequal dimensions in Sigma and EMcov.")
+      }
+    }
+
+    ## Check the number of groups by the size of the lists.
+    unequalGroups <- !all(length(rawData) == c(length(Sigma), length(Mu)))
+    if (unequalGroups) stop("Unequal number of groups in rawData, Sigma, Mu.
+       For multiple-group models, rawData must be a list of data frames,
+       NOT a single data frame with a \"group\" column.")
+
+    ## In each group, check Sigma is symmetric and dimensions match rawData and Mu.
+    for (g in seq_along(rawData)) {
+      if (!isSymmetric(Sigma[[g]])) stop("Sigma in group ", g, " not symmetric.")
+      unequalDim <- !all(ncol(rawData[[g]]) == c(nrow(Sigma[[g]]), length(Mu[[g]])))
+      if (unequalDim) stop("Unequal dimensions in rawData, Sigma, Mu.")
+    }
+
+    ## Check for names of group levels. If NULL, assign arbitrary ones.
+    if (!hasArg(group)) group <- "group"
+    if (!is.character(group)) stop("The \"group\" argument must be a character string.")
+    if (is.null(names(rawData))) {
+      group.label <- paste0("g", seq_along(rawData))
+    } else {
+      group.label <- names(rawData)
+    }
+  }
+
+  ## save a copy as myTransDat, whose elements will be replaced iteratively by
+  ## group and by missing data pattern within group.
+  myTransDat <- rawData
+  names(myTransDat) <- group.label
+  output <- list()
+
+  #########################
+  ## Data Transformation ##
+  #########################
+
+  for (g in seq_along(group.label)) {
+    if (transformation == 1) {
+      ## get missing data patterns
+      R <- ifelse(is.na(rawData[[g]]), 1, 0)
+      rowMissPatt <- apply(R, 1, function(x) paste(x, collapse = ""))
+      patt <- unique(rowMissPatt)
+      myRows <- lapply(patt, function(x) which(rowMissPatt == x))
+
+      ## for each pattern, apply transformation
+      tStart <- Sys.time()
+      transDatList <- lapply(patt, trans1, rowMissPatt = rowMissPatt,
+                             dat = rawData[[g]], Sigma = Sigma[[g]], Mu = Mu[[g]])
+      output$timeTrans <- Sys.time() - tStart
+      for (i in seq_along(patt)) myTransDat[[g]][myRows[[i]], ] <- transDatList[[i]]
+    } else {
+      tStart <- Sys.time()
+      myTransDat[[g]] <- SavaleiYuan(dat = rawData[[g]], Sigma = Sigma[[g]],
+                                     Mu = Mu[[g]], EMcov = EMcov[[g]])
+      output$timeTrans <- Sys.time() - tStart
+    }
+  }
+
+  ## option to end function here
+  if (transDataOnly) {
+    for (g in seq_along(myTransDat)) myTransDat[[g]][ , group] <- group.label[g]
+    ## option to write transformed data to a file
+    if (writeTransData) {
+      ## Set a few options, if the user didn't.
+      if (!hasArg(writeArgs)) writeArgs <- list(file = "transformedData.dat",
+                                                row.names = FALSE, na = "-999")
+      if (!exists("file", where = writeArgs)) writeTransArgs$file <- "transformedData.dat"
+      if (!exists("row.names", where = writeArgs)) writeArgs$row.names <- FALSE
+      if (!exists("na", where = writeArgs)) writeArgs$na <- "-999"
+
+      ## add grouping variable and bind together into one data frame
+      for (g in seq_along(myTransDat)) myTransDat[[g]][ , group] <- group.label[g]
+      writeArgs$x <- do.call("rbind", myTransDat)
+
+      ## write to file, print details to screen
+      do.call("write.table", writeArgs)
+      cat("Transformed data was written to file \"", writeArgs$file, "\" in:\n\n",
+          getwd(), "\n\nunless path specified by user in 'file' argument.\n", sep = "")
+      return(invisible(writeArgs$x))
+    }
+    return(do.call("rbind", myTransDat))
+  }
+
+  #############################################
+  ## Bootstrap distribution of fit statistic ##
+  #############################################
+
+  ## draw bootstrap samples
+  if (!is.null(seed)) set.seed(seed)
+  bootSamples <- lapply(1:nBoot, function(x) getBootSample(myTransDat, group, group.label))
+
+  ## option to write bootstrapped samples to file(s)
+  if (writeBootData) {
+    ## Set a few options, if the user didn't.
+    if (!hasArg(writeArgs)) writeArgs <- list(file = "bootstrappedSamples.dat",
+                                              row.names = FALSE, na = "-999")
+    if (!exists("file", where = writeArgs)) writeTransArgs$file <- "bootstrappedSamples.dat"
+    if (!exists("row.names", where = writeArgs)) writeArgs$row.names <- FALSE
+    if (!exists("na", where = writeArgs)) writeArgs$na <- "-999"
+
+    ## add indicator for bootstrapped sample, bind together into one data frame
+    for (b in seq_along(bootSamples)) bootSamples[[b]]$bootSample <- b
+    writeArgs$x <- do.call("rbind", bootSamples)
+
+    ## write to file, print details to screen
+    do.call("write.table", writeArgs)
+    cat("Bootstrapped samples written to file \"", writeArgs$file, "\" in:\n\n",
+        getwd(), "\n\nunless path specified by user in 'file' argument.\n", sep = "")
+    return(invisible(bootSamples))
+  }
+
+  ## option to end function here
+  if (bootSamplesOnly) return(bootSamples)
+
+  ## check for lavaan arguments in (...)
+  lavaanArgs <- list(...)
+  lavaanArgs$group <- group
+
+  ## fit model to bootstrap samples, save distribution of chi-squared test stat
+  if (hasArg(x)) {
+    ## grab defaults from lavaan object "x"
+    lavaanArgs$slotParTable <- as.list(parTable(x))
+    lavaanArgs$slotModel <- x@Model
+    lavaanArgs$slotOptions <- lavInspect(x, "options")
+  } else {
+    lavaanArgs$model <- model
+    lavaanArgs$missing <- "fiml"
+    ## set defaults that will be necessary for many models to run, that will
+    ## probably not be specified explictly or included in lavaan syntax
+    lavaanArgs$meanstructure <- TRUE
+    if (!exists("auto.var", where = lavaanArgs)) lavaanArgs$auto.var <- TRUE
+    if (!exists("auto.cov.y", where = lavaanArgs)) lavaanArgs$auto.cov.y <- TRUE
+    if (!exists("auto.cov.lv.x", where = lavaanArgs)) lavaanArgs$auto.cov.lv.x <- TRUE
+  }
+  ## run bootstrap fits
+  if (showProgress) {
+    mypb <- utils::txtProgressBar(min = 1, max = nBoot, initial = 1, char = "=",
+                                  width = 50, style = 3, file = "")
+    bootFits <- numeric()
+    tStart <- Sys.time()
+    for (j in 1:nBoot) {
+      bootFits[j] <- fitBootSample(bootSamples[[j]], args = lavaanArgs,
+                                   suppress = suppressWarn)
+      utils::setTxtProgressBar(mypb, j)
+    }
+    close(mypb)
+    output$timeFit <- Sys.time() - tStart
+  } else {
+    tStart <- Sys.time()
+    bootFits <- sapply(bootSamples, fitBootSample, args = lavaanArgs,
+                       suppress = suppressWarn)
+    output$timeFit <- Sys.time() - tStart
+  }
+
+  ## stack groups, save transformed data and distribution in output object
+  for (g in seq_along(myTransDat)) myTransDat[[g]][ , group] <- group.label[g]
+  output$Transformed.Data <- do.call("rbind", myTransDat)
+  output$Bootstrapped.Distribution <- bootFits
+  output$Original.ChiSquared <- ChiSquared
+  if (hasArg(x)) {
+    output$Degrees.Freedom <- lavInspect(x, "fit")["df"]
+  } else {
+    convSamp <- which(!is.na(bootFits))[1]
+    lavaanArgs$data <- bootSamples[[convSamp]]
+	lavaanlavaan <- function(...) { lavaan::lavaan(...) }
+    output$Degrees.Freedom <- lavInspect(do.call(lavaanlavaan, lavaanArgs), "fit")["df"]
+  }
+
+  ## calculate bootstrapped p-value
+  output$Bootstrapped.p.Value <- mean(bootFits >= ChiSquared, na.rm = TRUE)
+
+  ## print warning if any models didn't converge
+  if (any(is.na(bootFits))) {
+    nonConvMessage <- paste("Model did not converge for the following bootstrapped samples",
+                            paste(which(is.na(bootFits)), collapse = "\t"), sep = ":\n")
+    warning(nonConvMessage)
+  }
+
+  finalResult <- new("BootMiss", time = list(transform = output$timeTrans, fit = output$timeFit), transData = output$Transformed.Data, bootDist = output$Bootstrapped.Distribution, origChi = output$Original.ChiSquared, df = output$Degrees.Freedom, bootP = output$Bootstrapped.p.Value)
+
+  finalResult
+}
+
+
+## ----------------
+## Hidden Functions
+## ----------------
+
 ## Function to execute Transformation 1 on a single missing-data pattern
 trans1 <- function(MDpattern, rowMissPatt, dat, Sigma, Mu) {
   myRows <- which(rowMissPatt == MDpattern)
@@ -151,7 +605,7 @@ trans2 <- function(dat, Sigma, Mu, EMcov) {
   Mjs <- vector("list", J)
 
   ## Create Duplication Matrix and its inverse (Magnus & Neudecker, 1999)
-  Dup <- lavaan::duplicationMatrix(p)
+  Dup <- lavaan::lav_matrix_duplication(p)
   Dupinv <- solve(t(Dup) %*% Dup) %*% t(Dup)
 
   ## step through each MD pattern, populate Hjs and Mjs
@@ -338,6 +792,7 @@ getBootSample <- function(groupDat, group, group.label) {
 }
 
 ## fit the model to a single bootstrapped sample and return chi-squared
+#' @importFrom lavaan lavInspect
 fitBootSample <- function(dat, args, suppress) {
   args$data <- dat
   lavaanlavaan <- function(...) { lavaan::lavaan(...) }
@@ -347,275 +802,11 @@ fitBootSample <- function(dat, args, suppress) {
     fit <- do.call(lavaanlavaan, args)
   }
   if (!exists("fit")) return(c(chisq = NA))
-  if (inspect(fit, "converged")) {
-    chisq <- inspect(fit, "fit")[c("chisq", "chisq.scaled")]
+  if (lavInspect(fit, "converged")) {
+    chisq <- lavInspect(fit, "fit")[c("chisq", "chisq.scaled")]
   } else {
     chisq <- NA
   }
   if (is.na(chisq[2])) return(chisq[1]) else return(chisq[2])
 }
 
-
-
-## overall function to apply any of the above functions
-bsBootMiss <- function(x, transformation = 2, nBoot = 500, model, rawData,
-                             Sigma, Mu, group, ChiSquared, EMcov,
-                             writeTransData = FALSE, transDataOnly = FALSE,
-                             writeBootData = FALSE, bootSamplesOnly = FALSE,
-                             writeArgs, seed = NULL, suppressWarn = TRUE,
-                             showProgress = TRUE, ...) {
-  if(writeTransData) transDataOnly <- TRUE
-  if(writeBootData) bootSamplesOnly <- TRUE
-
-  check.nBoot <- (!is.numeric(nBoot) | nBoot < 1L) & !transDataOnly
-  if (check.nBoot) stop("The \"nBoot\" argument must be a positive integer.")
-
-  ## Which transformation?
-  if (!(transformation %in% 1:2)) stop("User must specify transformation 1 or 2.
-       Consult Savalei & Yuan (2009) for advice.
-       Transformation 3 is not currently available.")
-  if (transformation == 2) SavaleiYuan <- trans2
-  #if (transformation == 3) SavaleiYuan <- trans3
-
-  ######################
-  ## Data Preparation ##
-  ######################
-
-  ## If a lavaan object is supplied, the extracted values for rawData, Sigma, Mu,
-  ## EMcov, and EMmeans will override any user-supplied arguments.
-  if (hasArg(x)) {
-    rawData <- lapply(x@Data@X, as.data.frame)
-    for (g in seq_along(rawData)) colnames(rawData[[g]]) <- x@pta$vnames$ov[[g]]
-    ChiSquared <- inspect(x, "fit")[c("chisq", "chisq.scaled")]
-    ChiSquared <- ifelse(is.na(ChiSquared[2]), ChiSquared[1], ChiSquared[2])
-    group <- x@Data@group
-    if (length(group) == 0) group <- "group"
-    group.label <- x@Data@group.label
-    if (length(group.label) == 0) group.label <- 1
-    Sigma <- x@Fit@Sigma.hat
-    Mu <- x@Fit@Mu.hat
-    EMcov <- x@SampleStats@cov
-  } else {
-  ## If no lavaan object is supplied, check that required arguments are.
-    suppliedData <- c(hasArg(rawData), hasArg(Sigma), hasArg(Mu))
-    if (!all(suppliedData)) {
-      stop("Without a lavaan fitted object, user must supply raw data and model-implied moments.")
-    }
-    if (!hasArg(model) & !(transDataOnly | bootSamplesOnly)) {
-      stop("Without model syntax or fitted lavaan object, user can only
-       call this function to save transformed data or bootstrapped samples.")
-    }
-    if (!hasArg(ChiSquared) & !(transDataOnly | bootSamplesOnly)) {
-      stop("Without a fitted lavaan object or ChiSquared argument,
-       user can only call this function to save transformed data,
-       bootstrapped samples, or bootstrapped chi-squared values.")
-    }
-    if (!any(c(transDataOnly, bootSamplesOnly))) {
-      if (!is.numeric(ChiSquared)) stop("The \"ChiSquared\" argument must be numeric.")
-    }
-
-    ## If user supplies one-group data & moments, convert to lists.
-    if (class(rawData) == "data.frame") {
-      rawData <- list(rawData)
-    }
-    if (class(rawData) != "list") {
-      stop("The \"rawData\" argument must be a data.frame or list of data frames.")
-    } else {
-      if (!all(sapply(rawData, is.data.frame))) stop("Every element of \"rawData\" must be a data.frame")
-    }
-    if (class(Sigma) == "matrix") Sigma <- list(Sigma)
-    if (is.numeric(Mu)) Mu <- list(Mu)
-
-    ## check whether EMcov was supplied for starting values in Trans2/Trans3
-    if (!hasArg(EMcov)) {
-      EMcov <- vector("list", length(Sigma))
-    } else {
-      if (class(EMcov) == "matrix") EMcov <- list(EMcov)
-    ## check EMcov is symmetric and dimensions match Sigma
-      for (g in seq_along(EMcov)) {
-        if (!isSymmetric(EMcov[[g]])) stop("EMcov in group ", g, " not symmetric.")
-        unequalDim <- !all(dim(EMcov[[g]]) == dim(Sigma[[g]]))
-        if (unequalDim) stop("Unequal dimensions in Sigma and EMcov.")
-      }
-    }
-
-    ## Check the number of groups by the size of the lists.
-    unequalGroups <- !all(length(rawData) == c(length(Sigma), length(Mu)))
-    if (unequalGroups) stop("Unequal number of groups in rawData, Sigma, Mu.
-       For multiple-group models, rawData must be a list of data frames,
-       NOT a single data frame with a \"group\" column.")
-
-    ## In each group, check Sigma is symmetric and dimensions match rawData and Mu.
-    for (g in seq_along(rawData)) {
-      if (!isSymmetric(Sigma[[g]])) stop("Sigma in group ", g, " not symmetric.")
-      unequalDim <- !all(ncol(rawData[[g]]) == c(nrow(Sigma[[g]]), length(Mu[[g]])))
-      if (unequalDim) stop("Unequal dimensions in rawData, Sigma, Mu.")
-    }
-
-    ## Check for names of group levels. If NULL, assign arbitrary ones.
-    if (!hasArg(group)) group <- "group"
-    if (!is.character(group)) stop("The \"group\" argument must be a character string.")
-    if (is.null(names(rawData))) {
-      group.label <- paste0("g", seq_along(rawData))
-    } else {
-      group.label <- names(rawData)
-    }
-  }
-
-  ## save a copy as myTransDat, whose elements will be replaced iteratively by
-  ## group and by missing data pattern within group.
-  myTransDat <- rawData
-  names(myTransDat) <- group.label
-  output <- list()
-
-  #########################
-  ## Data Transformation ##
-  #########################
-
-  for (g in seq_along(group.label)) {
-    if (transformation == 1) {
-      ## get missing data patterns
-      R <- ifelse(is.na(rawData[[g]]), 1, 0)
-      rowMissPatt <- apply(R, 1, function(x) paste(x, collapse = ""))
-      patt <- unique(rowMissPatt)
-      myRows <- lapply(patt, function(x) which(rowMissPatt == x))
-
-      ## for each pattern, apply transformation
-      tStart <- Sys.time()
-      transDatList <- lapply(patt, trans1, rowMissPatt = rowMissPatt,
-                             dat = rawData[[g]], Sigma = Sigma[[g]], Mu = Mu[[g]])
-      output$timeTrans <- Sys.time() - tStart
-      for (i in seq_along(patt)) myTransDat[[g]][myRows[[i]], ] <- transDatList[[i]]
-    } else {
-      tStart <- Sys.time()
-      myTransDat[[g]] <- SavaleiYuan(dat = rawData[[g]],vSigma = Sigma[[g]],
-                                     Mu = Mu[[g]], EMcov = EMcov[[g]])
-      output$timeTrans <- Sys.time() - tStart
-    }
-  }
-
-  ## option to end function here
-  if (transDataOnly) {
-    for (g in seq_along(myTransDat)) myTransDat[[g]][ , group] <- group.label[g]
-    ## option to write transformed data to a file
-    if (writeTransData) {
-      ## Set a few options, if the user didn't.
-      if (!hasArg(writeArgs)) writeArgs <- list(file = "transformedData.dat",
-                                                row.names = FALSE, na = "-999")
-      if (!exists("file", where = writeArgs)) writeTransArgs$file <- "transformedData.dat"
-      if (!exists("row.names", where = writeArgs)) writeArgs$row.names <- FALSE
-      if (!exists("na", where = writeArgs)) writeArgs$na <- "-999"
-
-      ## add grouping variable and bind together into one data frame
-      for (g in seq_along(myTransDat)) myTransDat[[g]][ , group] <- group.label[g]
-      writeArgs$x <- do.call("rbind", myTransDat)
-
-      ## write to file, print details to screen
-      do.call("write.table", writeArgs)
-      cat("Transformed data was written to file \"", writeArgs$file, "\" in:\n\n",
-          getwd(), "\n\nunless path specified by user in 'file' argument.\n", sep = "")
-      return(invisible(writeArgs$x))
-    }
-    return(do.call("rbind", myTransDat))
-  }
-
-  #############################################
-  ## Bootstrap distribution of fit statistic ##
-  #############################################
-
-  ## draw bootstrap samples
-  if (!is.null(seed)) set.seed(seed)
-  bootSamples <- lapply(1:nBoot, function(x) getBootSample(myTransDat, group, group.label))
-
-  ## option to write bootstrapped samples to file(s)
-  if (writeBootData) {
-    ## Set a few options, if the user didn't.
-    if (!hasArg(writeArgs)) writeArgs <- list(file = "bootstrappedSamples.dat",
-                                              row.names = FALSE, na = "-999")
-    if (!exists("file", where = writeArgs)) writeTransArgs$file <- "bootstrappedSamples.dat"
-    if (!exists("row.names", where = writeArgs)) writeArgs$row.names <- FALSE
-    if (!exists("na", where = writeArgs)) writeArgs$na <- "-999"
-
-    ## add indicator for bootstrapped sample, bind together into one data frame
-    for (b in seq_along(bootSamples)) bootSamples[[b]]$bootSample <- b
-    writeArgs$x <- do.call("rbind", bootSamples)
-
-    ## write to file, print details to screen
-    do.call("write.table", writeArgs)
-    cat("Bootstrapped samples written to file \"", writeArgs$file, "\" in:\n\n",
-        getwd(), "\n\nunless path specified by user in 'file' argument.\n", sep = "")
-    return(invisible(bootSamples))
-  }
-
-  ## option to end function here
-  if (bootSamplesOnly) return(bootSamples)
-
-  ## check for lavaan arguments in (...)
-  lavaanArgs <- list(...)
-  lavaanArgs$group <- group
-
-  ## fit model to bootstrap samples, save distribution of chi-squared test stat
-  if (hasArg(x)) {
-    ## grab defaults from lavaan object "x"
-    lavaanArgs$slotParTable <- x@ParTable
-    lavaanArgs$slotModel <- x@Model
-    lavaanArgs$slotOptions <- x@Options
-  } else {
-    lavaanArgs$model <- model
-    lavaanArgs$missing <- "fiml"
-    ## set defaults that will be necessary for many models to run, that will
-    ## probably not be specified explictly or included in lavaan syntax
-    lavaanArgs$meanstructure <- TRUE
-    if (!exists("auto.var", where = lavaanArgs)) lavaanArgs$auto.var <- TRUE
-    if (!exists("auto.cov.y", where = lavaanArgs)) lavaanArgs$auto.cov.y <- TRUE
-    if (!exists("auto.cov.lv.x", where = lavaanArgs)) lavaanArgs$auto.cov.lv.x <- TRUE
-  }
-  ## run bootstrap fits
-  if (showProgress) {
-    mypb <- txtProgressBar(min = 1, max = nBoot, initial = 1, char = "=",
-                           width = 50, style = 3, file = "")
-    bootFits <- numeric()
-    tStart <- Sys.time()
-    for (j in 1:nBoot) {
-      bootFits[j] <- fitBootSample(bootSamples[[j]], args = lavaanArgs,
-                                   suppress = suppressWarn)
-      setTxtProgressBar(mypb, j)
-    }
-    close(mypb)
-    output$timeFit <- Sys.time() - tStart
-  } else {
-    tStart <- Sys.time()
-    bootFits <- sapply(bootSamples, fitBootSample, args = lavaanArgs,
-                       suppress = suppressWarn)
-    output$timeFit <- Sys.time() - tStart
-  }
-
-  ## stack groups, save transformed data and distribution in output object
-  for (g in seq_along(myTransDat)) myTransDat[[g]][ , group] <- group.label[g]
-  output$Transformed.Data <- do.call("rbind", myTransDat)
-  output$Bootstrapped.Distribution <- bootFits
-  output$Original.ChiSquared <- ChiSquared
-  if (hasArg(x)) {
-    output$Degrees.Freedom <- inspect(x, "fit")["df"]
-  } else {
-    convSamp <- which(!is.na(bootFits))[1]
-    lavaanArgs$data <- bootSamples[[convSamp]]
-	lavaanlavaan <- function(...) { lavaan::lavaan(...) }
-    output$Degrees.Freedom <- inspect(do.call(lavaanlavaan, lavaanArgs), "fit")["df"]
-  }
-
-  ## calculate bootstrapped p-value
-  output$Bootstrapped.p.Value <- mean(bootFits >= ChiSquared, na.rm = TRUE)
-
-  ## print warning if any models didn't converge
-  if (any(is.na(bootFits))) {
-    nonConvMessage <- paste("Model did not converge for the following bootstrapped samples",
-                            paste(which(is.na(bootFits)), collapse = "\t"), sep = ":\n")
-    warning(nonConvMessage)
-  }
-
-  finalResult <- new("BootMiss", time = list(transform = output$timeTrans, fit = output$timeFit), transData = output$Transformed.Data, bootDist = output$Bootstrapped.Distribution, origChi = output$Original.ChiSquared, df = output$Degrees.Freedom, bootP = output$Bootstrapped.p.Value)
-
-  finalResult
-}
-
